#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
对参数的相关梯度进行调整

Authors: PanXu
Date:    2020/08/21 09:34:00
"""

from torch.nn import Module
from torch.nn.utils import clip_grad_norm_

from .grad_rescaled import GradRescaled


class GradNormalize(GradRescaled):
    """
    对梯度 进行 normalize
    """

    def __init__(self, max_normalize: float):
        """
        初始化
        :param max_normalize: 最大的正则化值，一般会设置为 5 或者 10.
        """

        self._max_normalize = max_normalize

        assert self._max_normalize > 0., f"{self._max_normalize} 非法, 应该大于 0"

    def __call__(self, model: Module):
        clip_grad_norm_(model.parameters(), self._max_normalize)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
main tag 管理器

Authors: PanXu
Date:    2020/10/16 16:25:00
"""


class MainTagManager:
    """
    所有的 main tag 从这里获取
    """

    EPOCH_USED_TIME = "epoch_used_time"
    EPOCH_LOSS = "epoch_loss"
    EPOCH_METRIC = "epoch_metric"



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/05/25 17:04:00
"""

import os
import logging
import logging.handlers


def config(level: int = logging.DEBUG,
           log_file_path: str = None,
           log_mode: str = 'w') -> None:
    """
    this function is deprecated
    log配置
    :param level: log level
    :param log_file_path: log输出路径
    :param log_mode: w or a文件写入模式
    :return: None
    """
    if log_file_path is None:
        logging.basicConfig(
            level=level,
            format='%(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
            datefmt='%Y-%m-%D %H:%M:%S')
    else:
        logging.basicConfig(
            level=level,
            format='%(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
            datefmt='%Y-%m-%D %H:%M:%S',
            filename=log_file_path,
            filemode=log_mode)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
label decoder

Authors: PanXu
Date:    2020/07/05 10:19:00
"""

from .model_label_decoder import ModelLabelDecoder

from .label_decoder.label_decoder import LabelDecoder
from .label_decoder.sequence_label_decoder import SequenceLabelDecoder

from .label_index_decoder.label_index_decoder import LabelIndexDecoder
from .label_index_decoder.crf_label_index_decoder import CRFLabelIndexDecoder
from .label_index_decoder.sequence_max_label_index_decoder import SequenceMaxLabelIndexDecoder
from .label_index_decoder.max_label_index_decoder import MaxLabelIndexDecoder


#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
所有 callback 来这里添加

Authors: PanXu
Date:    2020/10/16 17:39:00
"""
from typing import Set

from easytext.trainer import Record
from easytext.trainer.trainer_callback import TrainerCallback


class TrainerCallbackComposite(TrainerCallback):
    """
    所有 trainer callback composite
    """

    def __init__(self):
        self._callbacks: Set[TrainerCallback] = set()

    def add_callback(self, callback: TrainerCallback) -> "TrainerCallbackComposite":
        self._callbacks.add(callback)
        return self

    def remove_callback(self, callback: TrainerCallback) -> "TrainerCallbackComposite":

        if callback in self._callbacks:
            self._callbacks.remove(callback)
        return self

    def on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        for callback in self._callbacks:
            callback.on_train_epoch_start(trainer=trainer,
                                          record=record)

    def on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        for callback in self._callbacks:
            callback.on_train_epoch_stop(trainer=trainer,
                                         record=record)

    def on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        for callback in self._callbacks:
            callback.on_evaluate_validation_epoch_start(trainer=trainer,
                                                        record=record)

    def on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        for callback in self._callbacks:
            callback.on_evaluate_validation_epoch_stop(trainer=trainer,
                                                       record=record)

    def on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        for callback in self._callbacks:
            callback.on_training_complete(trainer=trainer,
                                          record=record)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
Metric 跟踪器

Authors: panxu(panxu@baidu.com)
Date:    2020/05/28 09:01:00
"""
import json
from typing import Dict

from easytext.metrics import ModelTargetMetric
from easytext.utils.json_util import json2str


class MetricTrackerItem:
    """
    Metric Tracker Item
    """

    def __init__(self, epoch: int,
                 train_metric: Dict,
                 train_model_target_metric: ModelTargetMetric,
                 validation_metric: Dict,
                 validation_model_target_metric: ModelTargetMetric):
        """
        :param epoch: 当前 epoch
        :param train_metric: 当前 epoch的模型 metric
        :param train_model_target_metric: 当前模型的 target metric
        :param validation_metric: 当前 epoch的模型 metric
        :param validation_model_target_metric: 当前模型的 target metric
        """
        self.epoch = epoch
        self.train_metric = train_metric
        self.train_model_target_metric = train_model_target_metric
        self.validation_metric = validation_metric
        self.validation_model_target_metric = validation_model_target_metric


class MetricTracker:
    """
    Metric 跟踪器
    """

    def __init__(self, patient: int = None):
        self._best_epoch = None
        self.patient = patient

        # 会将每一个 epoch 产生的都记录下来
        self.metric_tracker_dict: Dict[int, MetricTrackerItem] = dict()

    def _calculate_best(self, epoch: int) -> "MetricTracker":
        """
        比较当前 epoch 与 best, 返回 best epoch
        :param epoch: 当前 epoch
        :return:
        """

        if self._best_epoch is None:
            self._best_epoch = epoch
        else:
            best_metric_value = self.metric_tracker_dict[self._best_epoch].validation_model_target_metric.value

            current_metric_value = self.metric_tracker_dict[epoch].validation_model_target_metric.value

            if current_metric_value > best_metric_value:
                self._best_epoch = epoch

        return self

    def add_metric(self,
                   epoch: int,
                   train_metric: Dict,
                   train_model_target_metric: ModelTargetMetric,
                   validation_metric: Dict,
                   validation_model_target_metric: ModelTargetMetric) -> "MetricTracker":
        """
        添加 metric
        :param epoch: 当前 epoch
        :param train_metric: 当前 epoch的模型 metric
        :param train_model_target_metric: 当前模型的 target metric
        :param validation_metric: 当前 epoch的模型 metric
        :param validation_model_target_metric: 当前模型的 target metric
        :return:
        """
        item = MetricTrackerItem(epoch=epoch,
                                 train_metric=train_metric,
                                 train_model_target_metric=train_model_target_metric,
                                 validation_metric=validation_metric,
                                 validation_model_target_metric=validation_model_target_metric)
        self.metric_tracker_dict[epoch] = item

        self._calculate_best(epoch)
        return self

    def best(self):
        """
        返回最好的
        :return:
        """
        if self._best_epoch is not None:
            return self.metric_tracker_dict[self._best_epoch]
        return None

    def __getitem__(self, epoch: int):
        return self.metric_tracker_dict[epoch]

    def __len__(self):
        return len(self.metric_tracker_dict)

    def save(self, file_path: str):
        """
        将 metric tracker 保存到文件
        :param file_path:
        :return:
        """
        with open(file_path, mode="w", encoding="utf-8") as f:
            f.write(f"{json2str(self, indent=2)}\n")
        return self

    @classmethod
    def from_file(cls, file_path: str):
        """
        从文件中载入  metric tracker
        :param file_path: 文件路径
        :return:
        """

        with open(file_path, mode="r", encoding="utf-8") as f:
            metric_tracker = cls()
            data_dict = json.load(f)
            metric_tracker.patient = data_dict["patient"]
            metric_tracker._best_epoch = data_dict["_best_epoch"]

            for epoch, item in data_dict["metric_tracker_dict"].items():
                epoch = int(epoch)
                train_metric = item["train_metric"]

                validation_metric = item["validation_metric"]

                train_model_target_metric = ModelTargetMetric(
                    metric_name=item["train_model_target_metric"]["_metric_name"],
                    metric_value=item["train_model_target_metric"]["_metric_value"])

                validation_model_target_metric = ModelTargetMetric(
                    metric_name=item["validation_model_target_metric"]["_metric_name"],
                    metric_value=item["validation_model_target_metric"]["_metric_value"])

                metric_tracker.metric_tracker_dict[epoch] = MetricTrackerItem(
                    epoch=item["epoch"],
                    train_metric=train_metric,
                    train_model_target_metric=train_model_target_metric,
                    validation_metric=validation_metric,
                    validation_model_target_metric=validation_model_target_metric)

            return metric_tracker

    def early_stopping(self, epoch: int) -> bool:
        """
        early stopping
        :return True: 停止训练; False: 不会停止训练:
        """
        is_stopping = False

        if self.patient is not None:
            is_stopping = (epoch - self._best_epoch) > self.patient

        return is_stopping



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
测试 acc metric

Authors: panxu(panxu@baidu.com)
Date:    2020/05/30 14:56:00
"""
import torch
from easytext.metrics import AccMetric

from easytext.tests import ASSERT


def test_acc_metric():

    # 对应的label是 [1, 1, 0, 1]
    logits = torch.tensor([[1., 2.], [3., 4.], [5, 4.], [3., 7.]], dtype=torch.float)
    prediction_labels = torch.argmax(logits, dim=-1)

    golden_labels = torch.tensor([0, 1, 1, 0], dtype=torch.long)

    acc_metric = AccMetric()

    expect = 1/4

    acc = acc_metric(prediction_labels=prediction_labels,
                     gold_labels=golden_labels,
                     mask=None)

    ASSERT.assertAlmostEqual(expect, acc[acc_metric.ACC])
    ASSERT.assertAlmostEqual(expect, acc_metric.metric[acc_metric.ACC])

    # 对应的label是 [0, 1, 0, 1]
    logits = torch.tensor([[3., 2.], [4., 6.], [5, 4.], [3., 7.]], dtype=torch.float)
    prediction_labels = torch.argmax(logits, dim=-1)

    golden_labels = torch.tensor([0, 1, 1, 0], dtype=torch.long)

    acc = acc_metric(prediction_labels=prediction_labels, gold_labels=golden_labels, mask=None)

    expect = 2 / 4
    ASSERT.assertAlmostEqual(expect, acc[acc_metric.ACC])

    # 下面的会计算将两次结果综合起来
    expect = (1+2)/(4+4)
    ASSERT.assertAlmostEqual(expect, acc_metric.metric[acc_metric.ACC])


def test_synchronized_data():
    """
    测试 from_synchronized_data 和 to_synchronized_data
    :return:
    """

    acc_metric = AccMetric()

    sync_data, op = acc_metric.to_synchronized_data()

    ASSERT.assertEqual((2,), sync_data.size())
    ASSERT.assertEqual(0, sync_data[0].item())
    ASSERT.assertEqual(0, sync_data[1].item())

    # 对应的label是 [1, 1, 0, 1]
    logits = torch.tensor([[1., 2.], [3., 4.], [5, 4.], [3., 7.]], dtype=torch.float)
    prediction_labels = torch.argmax(logits, dim=-1)

    golden_labels = torch.tensor([0, 1, 1, 0], dtype=torch.long)

    acc_metric(prediction_labels=prediction_labels, gold_labels=golden_labels, mask=None)

    # acc = 1/4

    sync_data, op = acc_metric.to_synchronized_data()
    ASSERT.assertListEqual([1, 4], sync_data.tolist())

    acc_metric.from_synchronized_data(sync_data=sync_data, reduce_op=op)
    acc = acc_metric.metric

    expect = 1/4
    ASSERT.assertAlmostEqual(expect, acc[AccMetric.ACC])

    new_sync_data, op = acc_metric.to_synchronized_data()

    ASSERT.assertListEqual(sync_data.tolist(), new_sync_data.tolist())










#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
label decoder

Authors: PanXu
Date:    2020/07/05 10:38:00
"""
from typing import List
import torch


class LabelDecoder:
    """
    将 label index 转化成 label，最终用户需要的结果
    """

    def __call__(self, label_indices: torch.LongTensor, mask: torch.ByteTensor) -> List:
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
loss 曲线的 callback

Authors: PanXu
Date:    2020/10/16 17:55:00
"""

import logging
import time

import torch
from torch.utils.tensorboard import SummaryWriter
from torch import distributed as Distributed

from easytext.trainer import Record
from easytext.trainer.trainer_callback import DistributedCallback
from easytext.trainer.trainer_callback import LossCallback
from easytext.trainer.tensorboard_manager import MainTagManager


class DistributedLossCallback(DistributedCallback):
    """
    用于分布式训练, loss callback 用来统计 epoch loss 变化
    """

    def __init__(self, tensorboard_summary_writer: SummaryWriter, target_rank: int = 0):
        """
        初始化化
        :param target_rank: 目标 rank 进程中做 loss 相关操作
        :param tensorboard_summary_writer: 对于非 target_rank 进程，该参数应该是 None, 会做校验
        """
        super().__init__(target_rank=target_rank)

        if self._is_target_rank():
            assert tensorboard_summary_writer is not None, f"在 {self._target_rank} 进程中, SummaryWriter 不能为 None."
        else:
            assert tensorboard_summary_writer is None, f"在 {self._target_rank} 进程中, SummaryWriter 必须设置为 None."

        self._loss_callback = LossCallback(tensorboard_summary_writer=tensorboard_summary_writer)

    def _on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        self._loss_callback.on_train_epoch_start(trainer=trainer, record=record)

    def _on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        self._loss_callback.on_train_epoch_stop(trainer=trainer, record=record)

    def on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        loss = torch.tensor(record.epoch_train_loss, dtype=torch.float)

        if Distributed.get_backend() == "nccl":
            loss.to(Distributed.get_rank())

        Distributed.reduce(tensor=loss, dst=self._target_rank)

    def _on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        self._loss_callback.on_evaluate_validation_epoch_start(trainer=trainer, record=record)

    def _on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        self._loss_callback.on_evaluate_validation_epoch_stop(trainer=trainer, record=record)

    def _on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        self._loss_callback.on_training_complete(trainer=trainer, record=record)






#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
tokenizer

Authors: panxu(panxu@baidu.com)
Date:    2020/05/13 16:17:00
"""

from typing import List
import unicodedata

from easytext.data.tokenizer.token import Token


class Tokenizer:
    """
    Tokenizer
    """

    def __init__(self, is_remove_invalidate_char: bool = False):
        """
        初始化
        :param is_remove_invalidate_char: 是否移除无效的字符。True: 移除; False: 不移除。
        无效字符包括: unicode: 0x0, 0xFFFD 以及 控制字符。一般来说，在分类任务中，应该设置为 True,
        但是在 序列标注任务中要小心，因为移除了会导致标注数据中的 token index 发生变化。
        这几个字符是非常特殊的字符，如果数据很干净，没有这些字符，那么，这个参数设置 True 还是 False
        都不会受到影响。
        """
        self._is_removed_invalidate_char = is_remove_invalidate_char

    def _is_whitespace(self, char):
        """
        Checks whether `chars` is a whitespace character.
        判断是否是 空白字符， 这包括 " ", "\t", "\n", and "\r".

        :param char:
        :return:
        """

        if char == " " or char == "\t" or char == "\n" or char == "\r":
            return True
        cat = unicodedata.category(char)
        if cat == "Zs":
            return True
        return False

    def _is_control(self, char):
        """
        检查是否是 控制字符。
        :param char: 字符
        :return:
        """
        """"""

        if char == "\t" or char == "\n" or char == "\r":
            return False
        cat = unicodedata.category(char)
        if cat.startswith("C"):
            return True
        return False

    def _clean_text(self, text):
        """
        将无效的字符以及空白符进行移除和替换
        :param text: 输入的文本
        :return:
        """
        output = []

        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xFFFD or self._is_control(char):
                if self._is_removed_invalidate_char:
                    continue

            output.append(char)

        return "".join(output)

    def tokenize(self, text: str) -> List[Token]:
        """
        tokenize
        :param text: 输入文本
        :return:
        """
        return self._tokenize(self._clean_text(text))

    def _tokenize(self, text: str) -> List[Token]:
        """
        实际进行 tokenize 的方法
        :param text:
        :return:
        """
        raise NotImplementedError

    def batch_tokenize(self, batch_text: List[str]) -> List[List[Token]]:
        """
        批量 tokenize
        :param batch_text:
        :return:
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
f1 metric 基类，具体的 f1 需要子类实现

Authors: panxu(panxu@baidu.com)
Date:    2020/06/16 09:08:00
"""

from typing import Dict, Set, List, Union, Tuple
from collections import OrderedDict

import torch
from torch import Tensor
from torch.distributed import ReduceOp
from torch import distributed as TorchDist


from easytext.metrics import Metric


class F1Metric(Metric):
    """
    f1 metric 基类，具体的 f1 需要子类实现，来完成
    _true_positives, _false_positives, _false_negatives
    三个字典的填充, 但是注意: 对于 *-overall 是会自动计算的，不需要
    指定计算。
    """

    PRECISION = "precision"
    RECALL = "recall"
    F1 = "f1"

    PRECISION_OVERALL = f"{PRECISION}-overall"
    RECALL_OVERALL = f"{RECALL}-overall"
    F1_OVERALL = f"{F1}-overall"

    def __init__(self, labels: List[str]) -> None:
        """
        初始化
        :param labels: 最终输出的 F1 的 label
        :param is_distributed: True: 分布式 metric; False: 非分布式 metric
        """
        super().__init__()
        self._labels = labels

        # 下面之所以是字典，是为了计算多个 label 的 f1
        self._true_positives: Dict[str, int] = OrderedDict()
        self._false_positives: Dict[str, int] = OrderedDict()
        self._false_negatives: Dict[str, int] = OrderedDict()

        for label in labels:
            self._true_positives[label] = 0
            self._false_positives[label] = 0
            self._false_negatives[label] = 0

    def __call__(self,
                 prediction_labels: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: torch.LongTensor) -> Dict:
        """
        计算 metric. 返回的是 F1 字典:

        {"precision_[tag]": [value],
         "recall_[tag]" : [value],
         "f1-measure_[tag]": [value],
         "precision-overall": [value],
         "recall-overall": [value],
         "f1-measure-overall": [value]}

         其中的 [tag] 是 实际返回的tag

        :param prediction_labels: 预测的结果, shape: (B, SeqLen)
        :param gold_labels: 实际的结果, shape: (B, SeqLen)
        :param mask: 对 predictions 和 gold label 的 mask, shape: (B, SeqLen)
        :return: 当前的 metric 计算字典结果.
        """

        raise NotImplementedError()

    def _metric(self,
                true_positives: Dict[str, int],
                false_positives: Dict[str, int],
                false_negatives: Dict[str, int]) -> Dict:
        """
        计算 metric, 注意是 输入是的字典
        :param true_positives:
        :param false_positives:
        :param false_negatives:
        :return:
        """

        all_tags: Set[str] = set()
        all_tags.update(true_positives.keys())
        all_tags.update(false_positives.keys())
        all_tags.update(false_negatives.keys())
        all_metrics = {}

        for tag in all_tags:
            precision, recall, f1_measure = self._compute_metrics(true_positives[tag],
                                                                  false_positives[tag],
                                                                  false_negatives[tag])
            precision_key = f"{F1Metric.PRECISION}-{tag}"
            recall_key = f"{F1Metric.RECALL}-{tag}"
            f1_key = f"{F1Metric.F1}-{tag}"
            all_metrics[precision_key] = precision
            all_metrics[recall_key] = recall
            all_metrics[f1_key] = f1_measure

        # Compute the precision, recall and f1 for overall
        precision, recall, f1_measure = self._compute_metrics(sum(true_positives.values()),
                                                              sum(false_positives.values()),
                                                              sum(false_negatives.values()))
        all_metrics[F1Metric.PRECISION_OVERALL] = precision
        all_metrics[F1Metric.RECALL_OVERALL] = recall
        all_metrics[F1Metric.F1_OVERALL] = f1_measure

        return all_metrics

    @property
    def metric(self) -> Dict:
        """
        :return

        一个包含所有metric的字典:
        {
            precision-[tag] : float
            recall-[tag] : float
            f1-[tag] : float
            ...
        }

        另外注意 precision-overall, recall-overall, f1-overall 是所有的综合指标, 这是非常有必要的
        因为有多个tag，那么模型的指标衡量需要一个综合指标来衡量。
        """

        return self._metric(true_positives=self._true_positives,
                            false_positives=self._false_positives,
                            false_negatives=self._false_negatives)

    @staticmethod
    def _compute_metrics(true_positives: int, false_positives: int, false_negatives: int):
        """
        计算 f1 metric
        """   
        precision = float(true_positives) / (float(true_positives + false_positives) + 1e-13)
        recall = float(true_positives) / (float(true_positives + false_negatives) + 1e-13)
        f1_measure = 2. * ((precision * recall) / (precision + recall + 1e-13))
        return precision, recall, f1_measure

    def reset(self):
        """
        将所有的状态reset, f1 重新计算。
        """
        for label in self._labels:
            self._true_positives[label] = 0
            self._false_positives[label] = 0
            self._false_negatives[label] = 0
        return self

    def to_synchronized_data(self) -> Tuple[Union[Dict[Union[str, int], Tensor], List[Tensor], Tensor], ReduceOp]:
        true_positives = torch.tensor([v for _, v in self._true_positives.items()], dtype=torch.long)

        false_positives = torch.tensor([v for _, v in self._false_positives.items()], dtype=torch.long)
        false_negatives = torch.tensor([v for _, v in self._false_negatives.items()], dtype=torch.long)

        sync_data = {"true_positives": true_positives,
                     "false_positives": false_positives,
                     "false_negatives": false_negatives}
        return sync_data, ReduceOp.SUM

    def from_synchronized_data(self, sync_data: Union[Dict[Union[str, int], Tensor], List[Tensor], Tensor],
                               reduce_op: ReduceOp) -> None:

        true_positives: Tensor = sync_data["true_positives"]
        false_positives = sync_data["false_positives"]
        false_negatives = sync_data["false_negatives"]

        assert true_positives.size(0) == len(self._true_positives), \
            f"true_positives length: {true_positives.size(0)} sync data " \
            f"与 self.true_positives length: {len(self._true_positives)} 不一致"

        assert false_positives.size(0) == len(self._false_positives), \
            f"true_positives length: {false_positives.size(0)} sync data " \
            f"与 self.true_positives length: {len(self._false_positives)} 不一致"

        assert false_negatives.size(0) == len(self._false_negatives), \
            f"true_positives length: {false_negatives.size(0)} sync data " \
            f"与 self.true_positives length: {len(self._false_negatives)} 不一致"

        for kv, sync_value in zip(self._true_positives.items(), true_positives):
            k, _ = kv
            self._true_positives[k] = sync_value

        for kv, sync_value in zip(self._false_positives.items(), false_positives):
            k, _ = kv
            self._false_positives[k] = sync_value

        for kv, sync_value in zip(self._false_negatives.items(), false_negatives):
            k, _ = kv
            self._false_negatives[k] = sync_value





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
预训练好的 词向量 载入其

Authors: panxu(panxu@baidu.com)
Date:    2020/06/24 11:35:00
"""

from typing import List, Dict


class PretrainedWordEmbeddingLoader:
    """
    预训练词向量 loader
    """

    def load(self) -> Dict[str, List[float]]:
        """
        载入预训练的词向量, 该函数是在初始化时候调用
        return: 返回词向量字典. 字典的 key 是 token, value 是 float List, 也就是向量内容;
        """
        raise NotImplementedError()

    @property
    def embedding_dim(self) -> int:
        """
        获取词向量维度
        return: 词向量的维度
        """
        raise NotImplementedError

    @property
    def embedding_dict(self) -> Dict[str, List[float]]:
        """
        embedding 字典,
        return: embedding dict, key 是 token, value: List[float] 向量，例如:
        {"the": [1.0, 2.0, ...]}
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
将模型输出转化成最终的label

Authors: PanXu
Date:    2020/07/05 10:20:00
"""
import torch
from typing import List, Tuple

from easytext.model import ModelOutputs
from easytext.data import LabelVocabulary


class ModelLabelDecoder:
    """
    模型 label 解码器，
    1. 在 predict 中直接使用, 将模型结果产生最终输出结果
    2. 在 metric 中使用 decode_label_index 将 logits 转化成 label index 用来进行 metric 计算
    """

    def decode_label_index(self, model_outputs: ModelOutputs) -> torch.LongTensor:
        """
        将模型输出解码成 index
        :param model_outputs: 模型输出
        :return:
        """
        raise NotImplementedError()

    def decode_label(self, model_outputs: ModelOutputs, label_indices: torch.LongTensor) -> List:
        """
        将 label index 解码成最终目标结果, 会使用到 label_vocabulary, 对 label index 进行实际的转换。
        对于分类来说, 假设 label index = [1, 5, 7], 返回结果是 labels=["军事", "政治", "经济"];
        对于其他任务来说也同样，返回的是用户能读懂的结果。
        :param model_outputs: 模型输出
        :param label_indices: decode_label_index 解码出的 label index
        :return: 转换后的最终结果
        """
        raise NotImplementedError()

    def __call__(self, model_outputs: ModelOutputs) -> List:
        """
        将模型输出解码成最终用户需要的输出
        :param model_outputs: 模型输出
        :return: 转化后的最终结果
        """
        label_indices = self.decode_label_index(model_outputs=model_outputs)
        return self.decode_label(model_outputs=model_outputs, label_indices=label_indices)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/05/18 10:37:00
"""
import torch
from easytext.model import ModelOutputs


class Loss:
    """
    Loss
    """

    def __call__(self, model_outputs: ModelOutputs, golden_label: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/05/16 00:59:00
"""

from typing import List, Dict, Tuple, Union

import torch
from torch import Tensor
from torch.distributed import ReduceOp

from easytext.model import ModelOutputs
from easytext.distributed import Synchronized


class Metric(Synchronized):
    """
    Metrics
    """

    def __call__(self, prediction_labels: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: torch.LongTensor) -> Dict:
        """
        返回metric的结果, 依然来说，因为 model_outputs 是一个bath产生的结果，
        所以这里返回的是一个 batch 的 metric 结果。
        :param prediction_labels: 预测的label 结果，一般是 label index，这与 gold_labels 是一致的
        :param gold_labels: 正确的结果
        :param mask: prediction 的 mask. 这里的 mask 类型是 Long, 不是 bool.
        :return:
        """
        raise NotImplementedError()

    @property
    def metric(self) -> Dict:
        """
        返回全部计算的 metric, 也就是将每一个 batch 计算的metric合并到一起的，总的 metric
        :return:
        """
        raise NotImplementedError()

    def reset(self):
        raise NotImplementedError()

    def to_synchronized_data(self) -> Tuple[Union[Dict[Union[str, int], Tensor], List[Tensor], Tensor], ReduceOp]:
        raise NotImplementedError()

    def from_synchronized_data(self, sync_data: Union[Dict[Union[str, int], Tensor], List[Tensor], Tensor],
                               reduce_op: ReduceOp) -> None:
        raise NotImplementedError()


class ModelTargetMetric:
    """
    模型的目标 metric 数据，作为模型唯一度量的metric， 因为 metric 本身会返回多个metric,
    但是模型需要一个唯一的 metric 作为评估以及 early stopping 等。
    """

    def __init__(self, metric_name: str, metric_value: float):
        self._metric_name = metric_name
        self._metric_value = metric_value

    @property
    def name(self):
        return self._metric_name

    @property
    def value(self):
        return self._metric_value


class ModelMetricAdapter(Synchronized):
    """
    模型的 metric 在 __call__ 的属于与 metric 不同,
    """

    def __call__(self,
                 model_outputs: ModelOutputs,
                 golden_labels: Tensor) -> Tuple[Dict, ModelTargetMetric]:
        """
        在每一个 batch 中 计算metric
        :param model_outputs:
        :param golden_labels:
        :return: 当前batch下的 metric 字典值, 以及 Model Target Metric.
        Model Target Metric 的含义是用来对该模型的唯一评判指标
        """
        raise NotImplementedError()

    @property
    def metric(self) -> Tuple[Dict, ModelTargetMetric]:
        """
        将所有 batch 的 metric 结果整合在一起计算的 metric.
        一般会在一个 epoch 结束调用这个函数来，得到这个 epoch 的 metric结果。
        如果每一个 batch 结束后调用，得到的是前面所有 batch 到现在的综合结果。
        :return: 1. Metric Dict: 作为模型最详细的metric，包括每一个细分的 metric, 例如 sequence label, 返回每一个label
        的 metric
                 2. Model Target Metric: 作为模型的唯一评估指标，训练时候会被用在 early stopping 中。
        """
        raise NotImplementedError()

    def reset(self) -> "ModelMetricAdapter":
        """
        所有数据重置, 一般是在每一个 epoch 的开始 做这个操作，将所有数据清零。
        :return: self
        """
        raise NotImplementedError()

    def to_synchronized_data(self) -> Tuple[Union[Dict[Union[str, int], Tensor], List[Tensor], Tensor], ReduceOp]:
        raise NotImplementedError()

    def from_synchronized_data(self, sync_data: Union[Dict[Union[str, int], Tensor], List[Tensor], Tensor],
                               reduce_op: ReduceOp) -> None:
        raise NotImplementedError()







#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
token

Authors: panxu(panxu@baidu.com)
Date:    2020/05/13 15:31:00
"""

class Token:

    def __init__(self, text: str):
        self.text = text


#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
component 内建的 key

Authors: PanXu
Date:    2020/10/27 19:57:00
"""


class ComponentBuiltinKey:
    """
    组件内建 key 会用在config的文件中
    """
    OBJECT = "__object__"
    TYPENAME = "__type__"
    NAME_SPACE = "__name_space__"
    EASYTEXT_NAME_SPACE = "__easytext__"
    IS_TRAINING = "is_training"




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
cuda 相关的工具

Authors: panxu(panxu@baidu.com)
Date:    2020/06/01 15:34:00
"""

import torch


def cuda(obj, cuda_device: torch.device):
    """
    将复杂对象全部移动到 cuda 中
    :param obj: obj
    :param cuda_device: cuda device
    :return:
    """
    if isinstance(obj, torch.Tensor):
        return obj.cuda(cuda_device)
    elif isinstance(obj, dict):
        return {key: cuda(value, cuda_device) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [cuda(item, cuda_device) for item in obj]
    elif isinstance(obj, tuple):
        return tuple([cuda(item, cuda_device) for item in obj])
    else:
        return obj



# coding:utf-8
"""
时间日期工具包
"""
import time

FORMATE_1 = '%Y-%m-%d %H:%M:%S'
FORMATE_2 = '%Y-%m-%d_%H:%M:%S'
FORMATE_3 = '%Y-%m-%d'
FORMATE_4 = '%H:%M:%S'
FORMATE_5 = '%Y%m%d'
FORMATE_6 = '%Y%m%d%H%M%S'
FORMATE_7 = '%Y%m%d_%H:%M:%S'
FORMATE_8 = '%Y%m%d%H0000'
FORMATE_9 = '%Y%m%d_%H%M%S'

def datetime2timestamp(cur_datetime):
    """
    日期转换成时间戳
    :param cur_datetime: data_time
    :return: 时间戳整数
    """

    return int(time.mktime(cur_datetime.timetuple()))



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
工厂初始化

Authors: PanXu
Date:    2020/10/27 14:53:00
"""

from .component_factory import ComponentFactory



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 time consume callback,

因为只能通过 tensorboard 查看，所以无法使用 assert

Authors: PanXu
Date:    2020/10/16 16:32:00
"""
import time
import os
import numpy as np

from easytext.utils import log_util

from easytext.trainer import Record
from easytext.trainer.trainer_callback import TimeConsumeCallback

log_util.config()


def test_time_consume_callback(summary_writer):

    callback = TimeConsumeCallback(tensorboard_summary_writer=summary_writer)

    for i in range(1, 10):
        record = Record()
        record.epoch = i

        callback.on_train_epoch_start(trainer=None, record=record)
        time.sleep(0.02)
        callback.on_train_epoch_stop(trainer=None, record=record)

        callback.on_evaluate_validation_epoch_start(trainer=None, record=record)
        time.sleep(0.01)
        callback.on_evaluate_validation_epoch_stop(trainer=None, record=record)

    summary_writer.flush()


#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
组件注册器

Authors: PanXu
Date:    2020/10/27 15:02:00
"""

from typing import Type

from easytext.component.register import Registry
from easytext.component.register.registy import T


class ComponentRegister:
    """
    Component 注册器，也是类装饰器
    """

    @classmethod
    def register(cls, name_space: str, typename: str = None, is_allowed_exist: bool = False) -> T:
        """
        用作在类或函数上的装饰器
        :param typename: 注册的类或者函数的类型名称, 如果为 None 或者 "", 那么, 将会默认使用类或者函数的名字作为名字,
                         在配置文件中直接使用类名字或者函数名字即可。
        :param name_space: 注册的类或者函数的 name space
        :param is_allowed_exist: True: 允许名字重复，那么，后面的名字会将前面的名字覆盖, 正常来讲不应该出现这样的设置;
                                 False: 不允许名字重复, 如果出现重复，自己定义的名字需要进行修改
        :return:
        """
        register = Registry()

        def add_to_registry(registered_class: Type[T]):

            name = typename or registered_class.__name__

            register.register_class(cls=registered_class,
                                    name_space=name_space,
                                    name=name,
                                    is_allowed_exist=is_allowed_exist)
            return registered_class

        return add_to_registry



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
训练的配置

Authors: PanXu
Date:    2020/10/29 10:01:00
"""
from typing import Dict
from collections import OrderedDict
import json
import copy

from easytext.component.factory import ComponentFactory


class Config:
    """
    训练器的配置
    """

    def __init__(self, is_training: bool, config_file_path: str, encoding: str = "utf-8"):

        with open(config_file_path, encoding=encoding) as f:
            self._raw_config: OrderedDict = json.load(f, object_pairs_hook=OrderedDict)
            self._is_training = is_training
            component_factory = ComponentFactory(self._is_training)
            self._config = component_factory.create(self._raw_config)

    def __str__(self):
        return json.dumps(self._raw_config, ensure_ascii=False)

    def __getattr__(self, key):
        return self._config[key]

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_config"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        component_factory = ComponentFactory(self._is_training)
        self._config = component_factory.create(self._raw_config.copy())




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/05/13 14:52:00
"""

from .model_outputs import ModelOutputs
from .model import Model
from .model_inputs import ModelInputs



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
初始化文件

Authors: PanXu
Date:    2020/10/20 17:24:00
"""

from .rnn_seq2seq import RnnSeq2Seq



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
基于 attention 的 seq2vec

Authors: PanXu
Date:    2020/07/18 12:00:00
"""
from typing import Optional

import torch
from torch.nn import Module
from torch.nn import Linear
from torch.nn import Tanh

from easytext.utils.nn.nn_util import masked_softmax


class AttentionSeq2Vec(Module):
    """
    基于 attention 将 seq2vec. 具体操作如下:

    1. sequence: (B, seq_len, input_size)
    2. K = WkSeqeunce 将 sequence 进行变换, K shape: (B, seq_len, query_hidden_size)
    3. Q = Shape: (query_hidden_size)
    4. attention = softmax(KQ), shape: (B, seq_len)
    5. V = WvSequence, shape: (B, seq_len, value_hidden_size); 如果 value_hidden_size is None,
    shape: (B, seq_len, input_size)
    6. sum(V*attention, dim=-1), shape: (B, input_size)
    """

    def __init__(self,
                 input_size: int,
                 query_hidden_size: int,
                 value_hidden_size: Optional[int] = None):
        """
        初始化。遵循 Q K V，计算 attention 方式。
        :param input_size: 输入的 sequence token 的 embedding dim
        :param query_hidden_size: 将 seqence 变成 Q 的时候，变换后的 token embedding dim.
        :param value_hidden_size: 将 seqence 变成 V 的时候, 变换后的 token embedding dim.
        如果 value_hidden_size is None, 那么，该模型就与 2016-Hierarchical Attention Networks for Document Classification
        是一致的, 最后的输出结果 shape (B, seq_len, input_size);
        如果 value_hidden_size 被设置了, 那么，就与 Attention is All your Need 中 变换是一致的, 最后的输出结果
        shape (B, seq_len, value_hidden_size)
        """
        super().__init__()
        # 对 seqence 转化成 K
        self.wk = Linear(in_features=input_size, out_features=query_hidden_size, bias=True)
        self.key_activation = Tanh()

        # attention 计算 dot 相似度，所以 bias = False
        self.attention = Linear(in_features=query_hidden_size, out_features=1, bias=False)

        self.wv = None
        if value_hidden_size is not None:
            self.wv = Linear(in_features=input_size, out_features=value_hidden_size, bias=True)

        self.reset_parameters()

    def reset_parameters(self):
        pass

    def forward(self, sequence: torch.LongTensor, mask: Optional[torch.ByteTensor]) -> torch.FloatTensor:
        """
        执行 attetion seq2vec
        :param sequence: 输入的token 序列, shape: (batch_size, seq_len, input_size)
        :param mask: mask shape: (batch_size, seq_len)
        :return: attention 编码向量, shape: (batch_size, value_hidden_size or input_size)
        """

        assert sequence.dim() == 3, f"sequence shape: (batch_size, seq_len, input_size)"

        if mask is not None:
            assert mask.dim() == 2, f"mask shape: (batch_size, seq_len)"

        # sequence 转换成 key, key shape: (B, seq_len, query_ebmedding_size)
        key = self.wk(sequence)
        key = self.key_activation(key)

        # Q * K 计算相似度, attention shape: (B, seq_len)
        attention = self.attention(key)
        attention = torch.squeeze(attention, dim=-1)

        # 对 attention 归一化
        if mask is not None:
            attention = masked_softmax(vector=attention, mask=mask)
        else:
            attention = torch.softmax(attention, dim=-1)

        # 计算value
        if self.wv is not None:
            # value shape: (B, seq_len, value_hidden_size)
            value = self.wv(sequence)
        else:
            # value shape: (B, seq_len, embedding_dim)
            value = sequence

        # attentioned_value shape: (B, seq_len, value_hidden_size or embedding_dim)
        attentioned_value = value * attention.unsqueeze(dim=-1)

        # 将 attenctioned value 相加在一起, vector shape: (B, value_hidden_size or embedding_dim)
        vector = torch.sum(attentioned_value, dim=1)
        return vector




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
loss 曲线的 callback

Authors: PanXu
Date:    2020/10/16 17:55:00
"""

import logging
import time

from torch.utils.tensorboard import SummaryWriter

from easytext.trainer import Record
from easytext.trainer.trainer_callback import TrainerCallback
from easytext.trainer.tensorboard_manager import MainTagManager


class LossCallback(TrainerCallback):
    """
    时间消耗 callback 用来统计不同阶段的时间, 时间单位是 毫秒 ms
    """

    def __init__(self, tensorboard_summary_writer: SummaryWriter):
        self._writer = tensorboard_summary_writer

    def on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        pass

    def on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        logging.info(f"Epoch {record.epoch} train loss: {record.epoch_train_loss:.4f}")

        self._writer.add_scalars(main_tag=MainTagManager.EPOCH_LOSS,
                                 tag_scalar_dict={"train": record.epoch_train_loss},
                                 global_step=record.epoch)

    def on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        pass

    def on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        logging.info(f"Epoch {record.epoch} evaluate validation loss: {record.epoch_validation_loss:.4f}")
        self._writer.add_scalars(main_tag=MainTagManager.EPOCH_LOSS,
                                 tag_scalar_dict={"validation": record.epoch_validation_loss},
                                 global_step=record.epoch)

    def on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        # 注意这里要 flush 不然可能会导致没有输出
        self._writer.flush()




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
metric callback

Authors: PanXu
Date:    2020/10/19 19:55:00
"""

import logging
import time

from torch.utils.tensorboard import SummaryWriter

from easytext.trainer import Record
from easytext.trainer.trainer_callback import TrainerCallback
from easytext.trainer.tensorboard_manager import MainTagManager


class MetricCallback(TrainerCallback):
    """
    绘制 train 和 evaluate metric 曲线
    """

    def __init__(self, tensorboard_summary_writer: SummaryWriter):
        self._writer = tensorboard_summary_writer

    def on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        pass

    def on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        self._writer.add_scalars(main_tag=MainTagManager.EPOCH_METRIC,
                                 tag_scalar_dict={"train": record.train_target_metric.value},
                                 global_step=record.epoch)

    def on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        pass

    def on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        self._writer.add_scalars(main_tag=MainTagManager.EPOCH_METRIC,
                                 tag_scalar_dict={"validation": record.validation_target_metric.value},
                                 global_step=record.epoch)

    def on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        # 注意这里要 flush 不然可能会导致没有输出
        self._writer.flush()




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
分布式接口

Authors: PanXu
Date:    2020/11/29 17:52:00
"""


class Distributed:
    """
    如果组件要兼容分布式和非分布式两种计算, 需要继承这个类
    在进行分布式或者多 GPU 会校验是否该接口子类
    """

    @property
    def is_distributed(self) -> bool:
        """
        当前是否是分布式/多GPU 运算
        :return: True: 是多GPU; False: 单 GPU 或 CPU
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 component factory

Authors: PanXu
Date:    2020/10/27 17:46:00
"""
import json
import os
from typing import Dict
from collections import OrderedDict

from torch.nn import Linear

from easytext.model import Model
from easytext.component import Component
from easytext.component.register import ComponentRegister
from easytext.component.factory import ComponentFactory
from easytext.component.register import Registry

from easytext.tests import ASSERT
from easytext.tests import ROOT_PATH


@ComponentRegister.register(typename="_MyModel", name_space="model")
class _MyModel(Model):

    def __init__(self, input_size: int, output_size: int):
        super().__init__()
        self.linear = Linear(in_features=input_size, out_features=output_size)


@ComponentRegister.register(typename="_NestedModel", name_space="model")
class _NestedModel(Model):

    def __init__(self, input_size: int, output_size: int, my_model: _MyModel):
        super().__init__()
        self.sub_model = Linear(in_features=input_size, out_features=output_size)
        self.my_model = my_model

    def reset_parameters(self):
        pass


@ComponentRegister.register(typename="_CustomerObj", name_space="test")
class _CustomerObj(Component):

    def __init__(self, is_training: bool, value: int):
        super().__init__(is_training=is_training)

        self.value = value


@ComponentRegister.register(typename="_ModelWithObjParam", name_space="model")
class _ModelWithObjParam(Model):

    def __init__(self, input_size: int, output_size: int, customer_obj: _CustomerObj):
        super().__init__()
        self.sub_model = Linear(in_features=input_size, out_features=output_size)
        self.customer_obj = customer_obj

    def reset_parameters(self):
        pass


@ComponentRegister.register(typename="_DictParamComponent", name_space="test")
class _DictParamComponent(Component):

    def __init__(self, is_training: bool, dict_value: Dict, customer_obj: _CustomerObj):
        super().__init__(is_training=is_training)
        self.dict_value = dict_value
        self.curstomer_obj = customer_obj


@ComponentRegister.register(typename="_TrainingComponent", name_space="training_component")
class _TrainingComponent(Component):
    """
    带有 trainging 的Component
    """

    def __init__(self, value: int, is_training: bool):
        super().__init__(is_training=is_training)

        if is_training:
            self.value = f"training_{value}"
        else:
            self.value = f"evaluate_{value}"


@ComponentRegister.register(typename="my_object", name_space="test")
def _my_object(value: str):
    return f"my_{value}"


@ComponentRegister.register(name_space="test")
class _DefaultTypename:

    def __init__(self, value):
        self.value = value + 1


def test_component_factory():
    Registry().clear_objects()

    model_json_file_path = "data/easytext/tests/component/model.json"
    model_json_file_path = os.path.join(ROOT_PATH, model_json_file_path)
    with open(model_json_file_path, encoding="utf-8") as f:
        config = json.load(f, object_pairs_hook=OrderedDict)

    factory = ComponentFactory(is_training=True)

    parserd_dict = factory.create(config=config)

    model = parserd_dict["model"]

    ASSERT.assertTrue(model.linear is not None)
    ASSERT.assertEqual((2, 4), (model.linear.in_features, model.linear.out_features))


def test_component_nested_factory():
    Registry().clear_objects()

    nested_json_file_path = "data/easytext/tests/component/nested.json"
    nested_json_file_path = os.path.join(ROOT_PATH, nested_json_file_path)
    with open(nested_json_file_path, encoding="utf-8") as f:

        param_dict = json.load(f, object_pairs_hook=OrderedDict)

    factory = ComponentFactory(is_training=True)

    parsed_dict = factory.create(config=param_dict)

    model = parsed_dict["model"]

    ASSERT.assertTrue(model.sub_model is not None)
    ASSERT.assertEqual((2, 4), (model.sub_model.in_features, model.sub_model.out_features))

    ASSERT.assertTrue(model.my_model is not None)
    ASSERT.assertTrue(model.my_model.linear is not None)
    ASSERT.assertEqual((3, 6), (model.my_model.linear.in_features, model.my_model.linear.out_features))


def test_component_training_factory():
    Registry().clear_objects()

    config_json_file_path = "data/easytext/tests/component/training.json"
    config_json_file_path = os.path.join(ROOT_PATH, config_json_file_path)
    with open(config_json_file_path, encoding="utf-8") as f:

        param_dict = json.load(f, object_pairs_hook=OrderedDict)

    factory = ComponentFactory(is_training=True)

    parsed_dict = factory.create(config=param_dict)

    my_component = parsed_dict["my_component"]

    ASSERT.assertEqual("training_3", my_component.value)


def test_component_evaluate_factory():
    Registry().clear_objects()

    config_json_file_path = "data/easytext/tests/component/training.json"
    config_json_file_path = os.path.join(ROOT_PATH, config_json_file_path)
    with open(config_json_file_path, encoding="utf-8") as f:

        param_dict = json.load(f, object_pairs_hook=OrderedDict)

    factory = ComponentFactory(is_training=False)

    parsed_dict = factory.create(config=param_dict)

    my_component = parsed_dict["my_component"]

    ASSERT.assertEqual("evaluate_3", my_component.value)


def test_component_with_object():
    """
    测试，当 component 构建的时候，某个参数是 object
    :return:
    """
    Registry().clear_objects()
    config_json_file_path = "data/easytext/tests/component/component_with_obj.json"
    config_json_file_path = os.path.join(ROOT_PATH, config_json_file_path)
    with open(config_json_file_path, encoding="utf-8") as f:
        param_dict = json.load(f, object_pairs_hook=OrderedDict)

    factory = ComponentFactory(is_training=False)

    parsed_dict = factory.create(config=param_dict)

    my_obj = parsed_dict["my_obj"]

    ASSERT.assertEqual(10, my_obj.value)

    my_component: _ModelWithObjParam = parsed_dict["my_component"]

    ASSERT.assertEqual(4, my_component.sub_model.in_features)
    ASSERT.assertEqual(2, my_component.sub_model.out_features)

    ASSERT.assertTrue(id(my_obj) == id(my_component.customer_obj))

    another_component: _ModelWithObjParam = parsed_dict["another_component"]

    ASSERT.assertTrue(id(my_component) != id(another_component))

    another_obj: _CustomerObj = parsed_dict["another_obj"]
    ASSERT.assertTrue(id(another_obj) == id(another_component.customer_obj))

    ASSERT.assertEqual(20, another_obj.value)

    dict_param_component: _DictParamComponent = parsed_dict["dict_param_component"]
    ASSERT.assertTrue(id(dict_param_component.curstomer_obj) == id(another_obj))

    ASSERT.assertEqual(1, dict_param_component.dict_value["a"])
    ASSERT.assertEqual(2, dict_param_component.dict_value["b"])
    ASSERT.assertEqual(30, dict_param_component.dict_value["c_obj"].value)

    my_object = parsed_dict["my_object"]
    ASSERT.assertEqual("my_test_value", my_object)


def test_default_typename():
    """
    测试，当 component 构建的时候，某个参数是 object
    :return:
    """
    Registry().clear_objects()
    config_json_file_path = "data/easytext/tests/component/default_typename.json"
    config_json_file_path = os.path.join(ROOT_PATH, config_json_file_path)
    with open(config_json_file_path, encoding="utf-8") as f:
        param_dict = json.load(f, object_pairs_hook=OrderedDict)

    factory = ComponentFactory(is_training=False)

    parsed_dict = factory.create(config=param_dict)

    default_typename = parsed_dict["default_typename"]

    ASSERT.assertEqual(10, default_typename.value)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
分布式使用的参数

Authors: PanXu
Date:    2020/12/08 20:34:00
"""

from easytext.component.register import EasytextRegister


@EasytextRegister.register()
class ProcessGroupParameter:
    """
    分布式训练初始化进程组使用的参数
    """

    def __init__(self, port: int, backend: str = "nccl", host: str = "127.0.0.1"):
        """
        分布式训练时候使用的相关参数
        :param port: 本机空闲的端口, 用作 init_process_group 中 init_method 参数使用
        :param backend: distributed 使用的 backend, "nccl" 或者 "gloo"
        :param host: 主机地址
        """
        self.backend = backend
        self.port = port
        self.host = host
        self.url = f"tcp://{self.host}:{self.port}"


@EasytextRegister.register()
class DistributedDataParallelParameter:
    """
    分布式训练使用的参数
    """

    def __init__(self, find_unused_parameters: bool):
        """
        DistributedDataParallel, 初始化用到的参数, 这里放置的是在训练中需要用户配置的。
        该类中没有的，会自动进行设置
        """
        self.find_unused_parameters = find_unused_parameters



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
同步数据

Authors: PanXu
Date:    2020/12/16 16:51:00
"""

from typing import Dict, Union

import torch
from torch import distributed as TorchDist
from torch.distributed import ReduceOp
from typing import List, Dict, Union, Tuple
from torch import Tensor


class Sync:
    """
    同步数据
    """

    @staticmethod
    def sync(value: Union[Tensor, int, float, Dict[str, Tensor]],
             device: torch.device,
             op: ReduceOp = ReduceOp.SUM) -> Union[Tensor, int, float, Dict[str, Tensor]]:
        """
        同步数据, in place 操作
        :param value: 需要同步的数据
        :param device: 当前 device
        :param op: op
        :return: 同步后的数据
        """
        if isinstance(value, int) or isinstance(value, float):
            return Sync.sync_value(value=value, device=device, op=op)
        elif isinstance(value, Tensor):
            return Sync.sync_tensor(value=value, device=device, op=op)
        elif isinstance(value, Dict):
            return Sync.sync_tensor_dict(value=value, device=device, op=op)
        else:
            raise RuntimeError(f"value 类型: {type(value)} 非法! 必须是 Tensor, int, float, Dict[str, Tensor] 中类型")


    @staticmethod
    def sync_tensor(value: Tensor,
                    device: torch.device,
                    op: ReduceOp = ReduceOp.SUM) -> Tensor:
        """
        同步 tensor
        :param value: tensor
        :param device: tensor 会放到指定 device 上进行同步, 这不是 tensor 本身带的 device
        :param op: reduce op
        :return: 返回同步后的 value, value 的 device 也会赋值成先前的
        """

        assert isinstance(value, Tensor), f"value: {type(value)} 类型必须是 Tensor"
        value_device = value.device

        if TorchDist.get_backend() == "nccl":
            assert device.type == "cuda", f"backend nccl, device 需要 cuda"

        sync_value = value.to(device)

        TorchDist.all_reduce(tensor=sync_value, op=op)

        return sync_value.to(value_device)

    @staticmethod
    def sync_value(value: Union[int, float],
                   device: torch.device,
                   op: ReduceOp = ReduceOp.SUM) -> Union[int, float]:
        """
        同步一个数值, 会将该数值先转换成 tensor 再进行同步
        :param value: 需要同步的数值
        :param device: 会放到指定 device 上进行同步
        :param op: reduce op
        :return: 同步后的数值，类型保持不变
        """

        if isinstance(value, float):
            sync_value_tensor = torch.tensor(value, dtype=torch.float, device=device)
        elif isinstance(value, int):
            sync_value_tensor = torch.tensor(value, dtype=torch.long, device=device)
        else:
            raise RuntimeError(f"value 字典，值必须是 float 或者 int 类型")

        sync_value_tensor = Sync.sync_tensor(value=sync_value_tensor, device=device, op=op)

        return sync_value_tensor.item()

    @staticmethod
    def sync_tensor_dict(value: Dict[str, Tensor],
                         device: torch.device,
                         op: ReduceOp = ReduceOp.SUM) -> Dict[str, Tensor]:
        """
        同步一个数值, 会将该数值先转换成 tensor 再进行同步
        :param value: 需要同步的数值
        :param device: 会放到指定 device 上进行同步
        :param op: reduce op
        :return: 同步后的数值，类型保持不变
        """
        for k, tensor in value.items():
            tensor = Sync.sync_tensor(value=tensor, device=device, op=op)

            value[k] = tensor

        return value



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
数据预处理 collate_fn

Authors: panxu(panxu@baidu.com)
Date:    2020/05/25 16:13:00
"""

from typing import Iterable, Dict

import torch

from .instance import Instance
from easytext.model import ModelInputs


class ModelCollate:
    """
    用在对数据处理产出模型的输入数据以及label, 是 torch.utils.data.DataLoader 中的 collate_fn 函数
    """

    def __call__(self, instances: Iterable[Instance]) -> ModelInputs:
        """
        collate_fn 执行
        :param instances: torch.utils.data.Dataset 中 __getitem__ 返回的是 Instance
        :return: 模型需要调用的数据
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
优化器创建工厂

Authors: panxu(panxu@baidu.com)
Date:    2020/06/01 16:11:00
"""
from torch.optim import Optimizer
from torch.optim import lr_scheduler
from easytext.model import Model


class OptimizerFactory:
    """
    Optimizer 创建工厂
    """

    def create(self, model: Model) -> "Optimizer":
        """
        创建 optimizer
        :param model:
        :return:
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 dynamic rnn

Authors: PanXu
Date:    2020/10/20 17:26:00
"""
import logging
import pytest

import torch
from torch.nn import LSTM, GRU, RNN

from easytext.modules import DynamicRnn, DynamicRnnOutput
from easytext.utils.json_util import json2str

from easytext.tests import ASSERT


@pytest.fixture(scope="package")
def sequence_embedding():
    sequence = torch.tensor([
        [
            [1, 2], [3, 4], [0, 0]
        ],

        [
            [9, 10], [0, 0], [0, 0]
        ],
        [
            [5, 6], [7, 8], [0, 0]
        ]], dtype=torch.float)

    mask = torch.tensor([
        [True, True, False],
        [True, False, False],
        [True, True, False]
    ], dtype=torch.bool)

    return sequence, mask


def test_dynamic_lstm(sequence_embedding):
    sequence, mask = sequence_embedding

    hidden_size = 4
    batch_size = 3
    sequence_len = 3

    lstm = LSTM(input_size=2,
                hidden_size=4,
                num_layers=2,
                batch_first=True,
                bidirectional=True)

    dynamic_rnn = DynamicRnn(rnn=lstm)

    rnn_output: DynamicRnnOutput = dynamic_rnn(sequence=sequence, mask=mask)

    logging.info(json2str(rnn_output))

    last_layer_h_n: torch.Tensor = rnn_output.last_layer_h_n

    last_layer_h_n_expect_size = (batch_size, hidden_size * 2)

    ASSERT.assertEqual(last_layer_h_n_expect_size, last_layer_h_n.size())

    last_layer_c_n: torch.Tensor = rnn_output.last_layer_c_n
    ASSERT.assertEqual(last_layer_h_n_expect_size, last_layer_c_n.size())

    sequence_encoding_expect_size = (batch_size, sequence_len, hidden_size * 2)
    senquence_encoding = rnn_output.output
    ASSERT.assertEqual(sequence_encoding_expect_size, senquence_encoding.size())


def test_dynamic_rnn(sequence_embedding):
    sequence, mask = sequence_embedding

    hidden_size = 4
    batch_size = 3
    sequence_len = 3

    rnn = RNN(input_size=2,
              hidden_size=4,
              num_layers=2,
              batch_first=True,
              bidirectional=True)

    dynamic_rnn = DynamicRnn(rnn=rnn)

    rnn_output: DynamicRnnOutput = dynamic_rnn(sequence=sequence, mask=mask)

    logging.info(json2str(rnn_output))

    last_layer_h_n: torch.Tensor = rnn_output.last_layer_h_n

    last_layer_h_n_expect_size = (batch_size, hidden_size * 2)

    ASSERT.assertEqual(last_layer_h_n_expect_size, last_layer_h_n.size())

    ASSERT.assertTrue(rnn_output.last_layer_c_n is None)

    sequence_encoding_expect_size = (batch_size, sequence_len, hidden_size * 2)
    senquence_encoding = rnn_output.output
    ASSERT.assertEqual(sequence_encoding_expect_size, senquence_encoding.size())



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
loss 曲线的 callback

Authors: PanXu
Date:    2020/10/16 17:55:00
"""

import logging
import time

from torch.utils.tensorboard import SummaryWriter
from torch import distributed as Distributed

from easytext.trainer import Record
from easytext.trainer.trainer_callback import TrainerCallback
from easytext.trainer.tensorboard_manager import MainTagManager


class DistributedCallback(TrainerCallback):
    """
    用于分布式训练中用到的 callback
    """

    def __init__(self, target_rank: int = 0):
        """
        初始化化
        :param target_rank: 目标 rank 进程中做 loss 相关操作
        """
        self._target_rank = target_rank

    def _is_target_rank(self) -> bool:
        """
        判断是否是目标 rank 进程
        :return: True: 是 target 进程; False: 不是 target 进程
        """
        return Distributed.get_rank() == self._target_rank

    def _on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        raise NotImplementedError()

    def on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        if self._is_target_rank():
            self._on_train_epoch_start(trainer=trainer, record=record)

    def _on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        raise NotImplementedError()

    def on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        if self._is_target_rank():
            return self._on_train_epoch_stop(trainer=trainer, record=record)

    def _on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        raise NotImplementedError()

    def on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        if self._is_target_rank():
            self._on_evaluate_validation_epoch_start(trainer=trainer, record=record)

    def _on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        raise NotImplementedError()

    def on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        if self._is_target_rank():
            self._on_evaluate_validation_epoch_stop(trainer=trainer, record=record)

    def _on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        raise NotImplementedError()

    def on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        if self._is_target_rank():
            self._on_training_complete(trainer=trainer, record=record)



